@@ -64,7 +64,8 @@
 #define D6_OPTION_UNICAST		12
 #define D6_OPTION_STATUS_CODE		13
 #define D6_OPTION_RAPID_COMMIT		14
-#define D6_OPTION_VENDOR		16
+#define D6_OPTION_VENDOR_CLASS		16
+#define D6_OPTION_VENDOR_OPTS		17
 #define D6_OPTION_SIP_SERVERS_NAME	21
 #define D6_OPTION_SIP_SERVERS_ADDRESS	22
 #define D6_OPTION_DNS_SERVERS		23
@@ -89,7 +90,6 @@
 
 #include "dhcp.h"
 #include "ipv6.h"
-extern const struct dhcp_opt dhcp6_opts[];
 
 struct dhcp6_message {
 	uint8_t type;
@@ -142,6 +142,10 @@ struct dhcp6_status {
 #define REC_MAX_RC		8
 #define HOP_COUNT_LIMIT		32
 
+/* RFC4242 3.1 */
+#define IRT_DEFAULT		86400
+#define IRT_MINIMUM		600
+
 #define DHCP6_RAND_MIN		-100
 #define DHCP6_RAND_MAX		100
 
@@ -218,6 +222,9 @@ struct dhcp6_state {
     ((const uint8_t *)(o) + sizeof(struct dhcp6_option))
 
 #ifdef INET6
+extern struct dhcp_opt *dhcp6_opts;
+extern size_t dhcp6_opts_len;
+
 void dhcp6_printoptions(void);
 int dhcp6_addrexists(const struct ipv6_addr *);
 int dhcp6_find_delegates(struct interface *);
