@@ -46,6 +46,7 @@
 #include "common.h"
 #include "dhcp.h"
 #include "dhcp6.h"
+#include "dhcpcd-embedded.h"
 #include "if-options.h"
 #include "ipv4.h"
 #include "platform.h"
@@ -70,6 +71,15 @@ unsigned long long options = 0;
 #define O_HOSTNAME_SHORT	O_BASE + 13
 #define O_DEV			O_BASE + 14
 #define O_NODEV			O_BASE + 15
+#define O_NOIPV4		O_BASE + 16
+#define O_NOIPV6		O_BASE + 17
+#define O_IAID			O_BASE + 18
+#define O_DEFINE		O_BASE + 19
+#define O_DEFINE6		O_BASE + 20
+#define O_EMBED			O_BASE + 21
+#define O_ENCAP			O_BASE + 22
+#define O_VENDOPT		O_BASE + 23
+#define O_VENDCLASS		O_BASE + 24
 
 char *dev_load;
 
@@ -129,13 +139,22 @@ const struct option cf_options[] = {
 	{"ipv6ra_own_default", no_argument,    NULL, O_IPV6RA_OWN_D},
 	{"ipv4only",        no_argument,       NULL, '4'},
 	{"ipv6only",        no_argument,       NULL, '6'},
+	{"noipv4",          no_argument,       NULL, O_NOIPV4},
+	{"noipv6",          no_argument,       NULL, O_NOIPV6},
 	{"noalias",         no_argument,       NULL, O_NOALIAS},
+	{"iaid",            required_argument, NULL, O_IAID},
 	{"ia_na",           no_argument,       NULL, O_IA_NA},
 	{"ia_ta",           no_argument,       NULL, O_IA_TA},
 	{"ia_pd",           no_argument,       NULL, O_IA_PD},
 	{"hostname_short",  no_argument,       NULL, O_HOSTNAME_SHORT},
 	{"dev",             required_argument, NULL, O_DEV},
 	{"nodev",           no_argument,       NULL, O_NODEV},
+	{"define",          required_argument, NULL, O_DEFINE},
+	{"define6",         required_argument, NULL, O_DEFINE6},
+	{"embed",           required_argument, NULL, O_EMBED},
+	{"encap",           required_argument, NULL, O_ENCAP},
+	{"vendopt",         required_argument, NULL, O_VENDOPT},
+	{"vendclass",       required_argument, NULL, O_VENDCLASS},
 	{NULL,              0,                 NULL, '\0'}
 };
 
@@ -257,12 +276,13 @@ parse_string_hwaddr(char *sbuf, ssize_t slen, const char *str, int clid)
 	/* If processing a string on the clientid, first byte should be
 	 * 0 to indicate a non hardware type */
 	if (clid && *str) {
-		*sbuf++ = 0;
+		if (sbuf)
+			*sbuf++ = 0;
 		l++;
 	}
 	c[3] = '\0';
 	while (*str) {
-		if (++l > slen) {
+		if (++l > slen && sbuf) {
 			errno = ENOBUFS;
 			return -1;
 		}
@@ -272,19 +292,23 @@ parse_string_hwaddr(char *sbuf, ssize_t slen, const char *str, int clid)
 			case '\0':
 				break;
 			case 'b':
-				*sbuf++ = '\b';
+				if (sbuf)
+					*sbuf++ = '\b';
 				str++;
 				break;
 			case 'n':
-				*sbuf++ = '\n';
+				if (sbuf)
+					*sbuf++ = '\n';
 				str++;
 				break;
 			case 'r':
-				*sbuf++ = '\r';
+				if (sbuf)
+					*sbuf++ = '\r';
 				str++;
 				break;
 			case 't':
-				*sbuf++ = '\t';
+				if (sbuf)
+					*sbuf++ = '\t';
 				str++;
 				break;
 			case 'x':
@@ -295,7 +319,7 @@ parse_string_hwaddr(char *sbuf, ssize_t slen, const char *str, int clid)
 						break;
 					c[i] = *str++;
 				}
-				if (c[1] != '\0') {
+				if (c[1] != '\0' && sbuf) {
 					c[2] = '\0';
 					*sbuf++ = strtol(c, NULL, 16);
 				} else
@@ -309,7 +333,7 @@ parse_string_hwaddr(char *sbuf, ssize_t slen, const char *str, int clid)
 						break;
 					c[i] = *str++;
 				}
-				if (c[2] != '\0') {
+				if (c[2] != '\0' && sbuf) {
 					i = strtol(c, NULL, 8);
 					if (i > 255)
 						i = 255;
@@ -318,18 +342,54 @@ parse_string_hwaddr(char *sbuf, ssize_t slen, const char *str, int clid)
 					l--;
 				break;
 			default:
-				*sbuf++ = *str++;
+				if (sbuf)
+					*sbuf++ = *str;
+				str++;
+				break;
 			}
-		} else
-			*sbuf++ = *str++;
+		} else {
+			if (sbuf)
+				*sbuf++ = *str;
+			str++;
+		}
 	}
 	if (punt_last) {
-		*--sbuf = '\0';
+		if (sbuf)
+			*--sbuf = '\0';
 		l--;
 	}
 	return l;
 }
 
+static int
+parse_iaid(uint8_t *iaid, const char *arg, size_t len)
+{
+	unsigned long l;
+	size_t s;
+	uint32_t u32;
+	char *np;
+
+	errno = 0;
+	l = strtoul(arg, &np, 0);
+	if (l <= (unsigned long)UINT32_MAX && errno == 0 && *np == '\0') {
+		u32 = htonl(l);
+		memcpy(iaid, &u32, sizeof(u32));
+		return 0;
+	}
+
+	if ((s = parse_string((char *)iaid, len, arg)) < 1) {
+		syslog(LOG_ERR, "%s: invalid IAID", arg);
+		return -1;
+	}
+	if (s < 4)
+		iaid[3] = '\0';
+	if (s < 3)
+		iaid[2] = '\0';
+	if (s < 2)
+		iaid[1] = '\0';
+	return 0;
+}
+
 static char **
 splitv(int *argc, char **argv, const char *arg)
 {
@@ -407,7 +467,7 @@ parse_addr(__unused struct in_addr *addr, __unused struct in_addr *net,
 #endif
 
 static const char *
-set_option_space(const char *arg, const struct dhcp_opt **d,
+set_option_space(const char *arg, const struct dhcp_opt **d, size_t *dl,
     struct if_options *ifo,
     uint8_t *request[], uint8_t *require[], uint8_t *no[])
 {
@@ -415,6 +475,7 @@ set_option_space(const char *arg, const struct dhcp_opt **d,
 #ifdef INET6
 	if (strncmp(arg, "dhcp6_", strlen("dhcp6_")) == 0) {
 		*d = dhcp6_opts;
+		*dl = dhcp6_opts_len;
 		*request = ifo->requestmask6;
 		*require = ifo->requiremask6;
 		*no = ifo->nomask6;
@@ -424,8 +485,10 @@ set_option_space(const char *arg, const struct dhcp_opt **d,
 
 #ifdef INET
 	*d = dhcp_opts;
+	*dl = dhcp_opts_len;
 #else
 	*d = NULL;
+	*dl = 0;
 #endif
 	*request = ifo->requestmask;
 	*require = ifo->requiremask;
@@ -433,10 +496,60 @@ set_option_space(const char *arg, const struct dhcp_opt **d,
 	return arg;
 }
 
+/* Pointer to last defined option */
+static struct dhcp_opt *ldop;
+static struct dhcp_opt *edop;
+
+void
+free_dhcp_opt_embenc(struct dhcp_opt *opt)
+{
+	size_t i;
+	struct dhcp_opt *o;
+
+	free(opt->var);
+
+	for (i = 0, o = opt->embopts; i < opt->embopts_len; i++, o++)
+		free(o->var);
+	free(opt->embopts);
+	opt->embopts_len = 0;
+	opt->embopts = NULL;
+
+	for (i = 0, o = opt->encopts; i < opt->encopts_len; i++, o++)
+		free(o->var);
+	free(opt->encopts);
+	opt->encopts_len = 0;
+	opt->encopts = NULL;
+}
+
+static char *
+strwhite(const char *s)
+{
+
+	while (*s != ' ' && *s != '\t') {
+		if (*s == '\0')
+			return NULL;
+		s++;
+	}
+	return UNCONST(s);
+}
+
+static char *
+strskipwhite(const char *s)
+{
+
+	while (*s == ' ' || *s == '\t') {
+		if (*s == '\0')
+			return NULL;
+		s++;
+	}
+	return UNCONST(s);
+}
+
 static int
 parse_option(struct if_options *ifo, int opt, const char *arg)
 {
-	int i;
+	int i, l, t;
+	unsigned int u;
 	char *p = NULL, *fp, *np, **nconf;
 	ssize_t s;
 	struct in_addr addr, addr2;
@@ -444,15 +557,18 @@ parse_option(struct if_options *ifo, int opt, const char *arg)
 	struct rt *rt;
 	const struct dhcp_opt *d;
 	uint8_t *request, *require, *no;
+	struct dhcp_opt **dop, *ndop;
+	size_t *dop_len, dl;
+	struct vivco *vivco;
 #ifdef INET6
-	long l;
-	uint32_t u32;
 	size_t sl;
-	struct if_iaid *iaid;
-	uint8_t _iaid[4];
+	struct if_ia *ia;
+	uint8_t iaid[4];
 	struct if_sla *sla, *slap;
 #endif
 
+	dop = NULL;
+	dop_len = NULL;
 	i = 0;
 	switch(opt) {
 	case 'f': /* FALLTHROUGH */
@@ -460,7 +576,8 @@ parse_option(struct if_options *ifo, int opt, const char *arg)
 	case 'n': /* FALLTHROUGH */
 	case 'x': /* FALLTHROUGH */
 	case 'T': /* FALLTHROUGH */
-	case 'U': /* We need to handle non interface options */
+	case 'U': /* FALLTHROUGH */
+	case 'V': /* We need to handle non interface options */
 		break;
 	case 'b':
 		ifo->options |= DHCPCD_BACKGROUND;
@@ -516,7 +633,7 @@ parse_option(struct if_options *ifo, int opt, const char *arg)
 			return -1;
 		}
 		errno = 0;
-		ifo->leasetime = (uint32_t)strtol(arg, NULL, 0);
+		ifo->leasetime = (uint32_t)strtoul(arg, NULL, 0);
 		if (errno == EINVAL || errno == ERANGE) {
 			syslog(LOG_ERR, "`%s' out of range", arg);
 			return -1;
@@ -530,8 +647,9 @@ parse_option(struct if_options *ifo, int opt, const char *arg)
 		}
 		break;
 	case 'o':
-		arg = set_option_space(arg, &d, ifo, &request, &require, &no);
-		if (make_option_mask(d, request, arg, 1) != 0) {
+		arg = set_option_space(arg, &d, &dl, ifo,
+		    &request, &require, &no);
+		if (make_option_mask(d, dl, request, arg, 1) != 0) {
 			syslog(LOG_ERR, "unknown option `%s'", arg);
 			return -1;
 		}
@@ -590,7 +708,7 @@ parse_option(struct if_options *ifo, int opt, const char *arg)
 	case 'v':
 		p = strchr(arg, ',');
 		if (!p || !p[1]) {
-			syslog(LOG_ERR, "invalid vendor format");
+			syslog(LOG_ERR, "invalid vendor format: %s", arg);
 			return -1;
 		}
 
@@ -614,22 +732,25 @@ parse_option(struct if_options *ifo, int opt, const char *arg)
 			ifo->vendor[0] = 0;
 		}
 
-		*p = '\0';
+		/* No need to strip the comma */
 		i = atoint(arg);
-		arg = p + 1;
 		if (i < 1 || i > 254) {
 			syslog(LOG_ERR, "vendor option should be between"
 			    " 1 and 254 inclusive");
 			return -1;
 		}
+
+		arg = p + 1;
 		s = VENDOR_MAX_LEN - ifo->vendor[0] - 2;
 		if (inet_aton(arg, &addr) == 1) {
 			if (s < 6) {
 				s = -1;
 				errno = ENOBUFS;
-			} else
+			} else {
 				memcpy(ifo->vendor + ifo->vendor[0] + 3,
 				    &addr.s_addr, sizeof(addr.s_addr));
+				s = sizeof(addr.s_addr);
+			}
 		} else {
 			s = parse_string((char *)ifo->vendor +
 			    ifo->vendor[0] + 3, s, arg);
@@ -740,19 +861,21 @@ parse_option(struct if_options *ifo, int opt, const char *arg)
 		ifo->options &= ~DHCPCD_IPV4LL;
 		break;
 	case 'O':
-		arg = set_option_space(arg, &d, ifo, &request, &require, &no);
-		if (make_option_mask(d, request, arg, -1) != 0 ||
-		    make_option_mask(d, require, arg, -1) != 0 ||
-		    make_option_mask(d, no, arg, 1) != 0)
+		arg = set_option_space(arg, &d, &dl, ifo,
+		    &request, &require, &no);
+		if (make_option_mask(d, dl, request, arg, -1) != 0 ||
+		    make_option_mask(d, dl, require, arg, -1) != 0 ||
+		    make_option_mask(d, dl, no, arg, 1) != 0)
 		{
 			syslog(LOG_ERR, "unknown option `%s'", arg);
 			return -1;
 		}
 		break;
 	case 'Q':
-		arg = set_option_space(arg, &d, ifo, &request, &require, &no);
-		if (make_option_mask(d, require, arg, 1) != 0 ||
-		    make_option_mask(d, request, arg, 1) != 0)
+		arg = set_option_space(arg, &d, &dl, ifo,
+		    &request, &require, &no);
+		if (make_option_mask(d, dl, require, arg, 1) != 0 ||
+		    make_option_mask(d, dl, request, arg, 1) != 0)
 		{
 			syslog(LOG_ERR, "unknown option `%s'", arg);
 			return -1;
@@ -781,14 +904,13 @@ parse_option(struct if_options *ifo, int opt, const char *arg)
 		    strncmp(arg, "classless_static_routes=", strlen("classless_static_routes=")) == 0 ||
 		    strncmp(arg, "ms_classless_static_routes=", strlen("ms_classless_static_routes=")) == 0)
 		{
-			fp = np = strchr(p, ' ');
+			fp = np = strwhite(p);
 			if (np == NULL) {
 				syslog(LOG_ERR, "all routes need a gateway");
 				return -1;
 			}
 			*np++ = '\0';
-			while (*np == ' ')
-				np++;
+			np = strskipwhite(np);
 			if (ifo->routes == NULL) {
 				ifo->routes = malloc(sizeof(*ifo->routes));
 				if (ifo->routes == NULL) {
@@ -909,21 +1031,37 @@ parse_option(struct if_options *ifo, int opt, const char *arg)
 		ifo->options &= ~DHCPCD_IPV4;
 		ifo->options |= DHCPCD_IPV6;
 		break;
+	case O_NOIPV4:
+		ifo->options &= ~DHCPCD_IPV4;
+		break;
+	case O_NOIPV6:
+		ifo->options &= ~DHCPCD_IPV6;
+		break;
 #ifdef INET
 	case O_ARPING:
-		if (parse_addr(&addr, NULL, arg) != 0)
-			return -1;
-		naddr = realloc(ifo->arping,
-		    sizeof(in_addr_t) * (ifo->arping_len + 1));
-		if (naddr == NULL) {
-			syslog(LOG_ERR, "%s: %m", __func__);
-			return -1;
+		while (arg && *arg != '\0') {
+			fp = strwhite(arg);
+			if (fp)
+				*fp++ = '\0';
+			if (parse_addr(&addr, NULL, arg) != 0)
+				return -1;
+			naddr = realloc(ifo->arping,
+			    sizeof(in_addr_t) * (ifo->arping_len + 1));
+			if (naddr == NULL) {
+				syslog(LOG_ERR, "%s: %m", __func__);
+				return -1;
+			}
+			ifo->arping = naddr;
+			ifo->arping[ifo->arping_len++] = addr.s_addr;
+			if (fp)
+				arg = strskipwhite(fp);
+			else
+				arg = NULL;
 		}
-		ifo->arping = naddr;
-		ifo->arping[ifo->arping_len++] = addr.s_addr;
 		break;
 	case O_DESTINATION:
-		if (make_option_mask(dhcp_opts, ifo->dstmask, arg, 2) != 0) {
+		if (make_option_mask(dhcp_opts, dhcp_opts_len,
+		    ifo->dstmask, arg, 2) != 0) {
 			if (errno == EINVAL)
 				syslog(LOG_ERR, "option `%s' does not take"
 				    " an IPv4 address", arg);
@@ -941,6 +1079,10 @@ parse_option(struct if_options *ifo, int opt, const char *arg)
 		}
 		break;
 #endif
+	case O_IAID:
+		if (parse_iaid(ifo->iaid, arg, sizeof(ifo->iaid)) == -1)
+			return -1;
+		ifo->options |= DHCPCD_IAID;
 	case O_IPV6RS:
 		ifo->options |= DHCPCD_IPV6RS;
 		break;
@@ -978,70 +1120,54 @@ parse_option(struct if_options *ifo, int opt, const char *arg)
 		ifo->ia_type = i;
 		if (arg == NULL)
 			break;
-		fp = strchr(arg, ' ');
+		fp = strwhite(arg);
 		if (fp)
 			*fp++ = '\0';
-		errno = 0;
-		l = strtol(arg, &np, 0);
-		if (l >= 0 && l <= (long)UINT32_MAX &&
-		    errno == 0 && *np == '\0')
-		{
-			u32 = htonl(l);
-			memcpy(&_iaid, &u32, sizeof(_iaid));
-			goto got_iaid;
-		}
-		if ((s = parse_string((char *)_iaid, sizeof(_iaid), arg)) < 1) {
-			syslog(LOG_ERR, "%s: invalid IAID", arg);
+		if (parse_iaid(iaid, arg, sizeof(iaid)) == -1)
 			return -1;
-		}
-		if (s < 4)
-			_iaid[3] = '\0';
-		if (s < 3)
-			_iaid[2] = '\0';
-		if (s < 2)
-			_iaid[1] = '\0';
-got_iaid:
-		iaid = NULL;
-		for (sl = 0; sl < ifo->iaid_len; sl++) {
-			if (ifo->iaid[sl].iaid[0] == _iaid[0] &&
-			    ifo->iaid[sl].iaid[1] == _iaid[1] &&
-			    ifo->iaid[sl].iaid[2] == _iaid[2] &&
-			    ifo->iaid[sl].iaid[3] == _iaid[3])
+		ia = NULL;
+		for (sl = 0; sl < ifo->ia_len; sl++) {
+			if (ifo->ia[sl].iaid[0] == iaid[0] &&
+			    ifo->ia[sl].iaid[1] == iaid[1] &&
+			    ifo->ia[sl].iaid[2] == iaid[2] &&
+			    ifo->ia[sl].iaid[3] == iaid[3])
 			{
-			        iaid = &ifo->iaid[sl];
+			        ia = &ifo->ia[sl];
 				break;
 			}
 		}
-		if (iaid == NULL) {
-			iaid = realloc(ifo->iaid,
-			    sizeof(*ifo->iaid) * (ifo->iaid_len + 1));
-			if (iaid == NULL) {
+		if (ia == NULL) {
+			ia = realloc(ifo->ia,
+			    sizeof(*ifo->ia) * (ifo->ia_len + 1));
+			if (ia == NULL) {
 				syslog(LOG_ERR, "%s: %m", __func__);
 				return -1;
 			}
-			ifo->iaid = iaid;
-			iaid = &ifo->iaid[ifo->iaid_len++];
-			iaid->iaid[0] = _iaid[0];
-			iaid->iaid[1] = _iaid[1];
-			iaid->iaid[2] = _iaid[2];
-			iaid->iaid[3] = _iaid[3];
-			iaid->sla = NULL;
-			iaid->sla_len = 0;
+			ifo->ia = ia;
+			ia = &ifo->ia[ifo->ia_len++];
+			ia->iaid[0] = iaid[0];
+			ia->iaid[1] = iaid[1];
+			ia->iaid[2] = iaid[2];
+			ia->iaid[3] = iaid[3];
+			ia->sla = NULL;
+			ia->sla_len = 0;
 		}
 		if (ifo->ia_type != D6_OPTION_IA_PD)
 			break;
 		for (p = fp; p; p = fp) {
-			fp = strchr(p, ' ');
-			if (fp)
+			fp = strwhite(p);
+			if (fp) {
 				*fp++ = '\0';
-			sla = realloc(iaid->sla,
-			    sizeof(*iaid->sla) * (iaid->sla_len + 1));
+				fp = strskipwhite(fp);
+			}
+			sla = realloc(ia->sla,
+			    sizeof(*ia->sla) * (ia->sla_len + 1));
 			if (sla == NULL) {
 				syslog(LOG_ERR, "%s: %m", __func__);
 				return -1;
 			}
-			iaid->sla = sla;
-			sla = &iaid->sla[iaid->sla_len++];
+			ia->sla = sla;
+			sla = &ia->sla[ia->sla_len++];
 			np = strchr(p, '/');
 			if (np)
 				*np++ = '\0';
@@ -1077,8 +1203,8 @@ got_iaid:
 				sla->sla_set = 0;
 				/* Sanity - check there are no more
 				 * unspecified SLA's */
-				for (sl = 0; sl < iaid->sla_len - 1; sl++) {
-					slap = &iaid->sla[sl];
+				for (sl = 0; sl < ia->sla_len - 1; sl++) {
+					slap = &ia->sla[sl];
 					if (slap->sla_set == 0 &&
 					    strcmp(slap->ifname, sla->ifname)
 					    == 0)
@@ -1088,14 +1214,14 @@ got_iaid:
 						    "same interface twice with "
 						    "an automatic SLA",
 						    sla->ifname);
-						iaid->sla_len--;
+						ia->sla_len--;
 						break;
 					}
 				}
 			}
 		}
-		break;
 #endif
+		break;
 	case O_HOSTNAME_SHORT:
 		ifo->options |= DHCPCD_HOSTNAME | DHCPCD_HOSTNAME_SHORT;
 		break;
@@ -1107,6 +1233,265 @@ got_iaid:
 	case O_NODEV:
 		ifo->options &= ~DHCPCD_DEV;
 		break;
+	case O_DEFINE:
+		dop = &ifo->dhcp_override;
+		dop_len = &ifo->dhcp_override_len;
+		/* FALLTHROUGH */
+	case O_DEFINE6:
+		if (dop == NULL) {
+			dop = &ifo->dhcp6_override;
+			dop_len = &ifo->dhcp6_override_len;
+		}
+		/* FALLTHROUGH */
+	case O_VENDOPT:
+		if (dop == NULL) {
+			dop = &ifo->vivso_override;
+			dop_len = &ifo->vivso_override_len;
+		}
+		edop = ldop = NULL;
+		/* FALLTHROUGH */
+	case O_EMBED:
+		if (dop == NULL) {
+			if (edop) {
+				dop = &edop->embopts;
+				dop_len = &edop->embopts_len;
+			} else if (ldop) {
+				dop = &ldop->embopts;
+				dop_len = &ldop->embopts_len;
+			} else {
+				syslog(LOG_ERR,
+				    "embed must be after a define or encap");
+				return -1;
+			}
+		}
+		/* FALLTHROUGH */
+	case O_ENCAP:
+		if (dop == NULL) {
+			if (ldop == NULL) {
+				syslog(LOG_ERR, "encap must be after a define");
+				return -1;
+			}
+			dop = &ldop->encopts;
+			dop_len = &ldop->encopts_len;
+		}
+
+		/* Shared code for define, define6, embed and encap */
+
+		/* code */
+		if (opt == O_EMBED) /* Embedded options don't have codes */
+			u = 0;
+		else {
+			fp = strwhite(arg);
+			if (!fp) {
+				syslog(LOG_ERR, "invalid syntax: %s", arg);
+				return -1;
+			}
+			*fp++ = '\0';
+			errno = 0;
+			u = strtoul(arg, &np, 0);
+			if (u > UINT32_MAX || errno != 0 || *np != '\0') {
+				syslog(LOG_ERR, "invalid code: %s", arg);
+				return -1;
+			}
+			arg = strskipwhite(fp);
+		}
+		/* type */
+		fp = strwhite(arg);
+		if (fp)
+			*fp++ = '\0';
+		np = strchr(arg, ':');
+		/* length */
+		if (np) {
+			*np++ = '\0';
+			if ((l = atoint(np)) == -1)
+				return -1;
+		} else
+			l = 0;
+		t = 0;
+		if (strcasecmp(arg, "request") == 0) {
+			t |= REQUEST;
+			arg = strskipwhite(fp);
+			fp = strwhite(arg);
+			if (fp == NULL) {
+				syslog(LOG_ERR, "incomplete request type");
+				return -1;
+			}
+			*fp++ = '\0';
+		} else if (strcasecmp(arg, "norequest") == 0) {
+			t |= NOREQ;
+			arg = strskipwhite(fp);
+			fp = strwhite(arg);
+			if (fp == NULL) {
+				syslog(LOG_ERR, "incomplete request type");
+				return -1;
+			}
+			*fp++ = '\0';
+		}
+		if (strcasecmp(arg, "index") == 0) {
+			t |= INDEX;
+			arg = strskipwhite(fp);
+			fp = strwhite(arg);
+			if (fp == NULL) {
+				syslog(LOG_ERR, "incomplete index type");
+				return -1;
+			}
+			*fp++ = '\0';
+		}
+		if (strcasecmp(arg, "array") == 0) {
+			t |= ARRAY;
+			arg = strskipwhite(fp);
+			fp = strwhite(arg);
+			if (fp == NULL) {
+				syslog(LOG_ERR, "incomplete array type");
+				return -1;
+			}
+			*fp++ = '\0';
+		}
+		if (strcasecmp(arg, "ipaddress") == 0)
+			t |= ADDRIPV4;
+		else if (strcasecmp(arg, "ip6address") == 0)
+			t |= ADDRIPV6;
+		else if (strcasecmp(arg, "string") == 0)
+			t |= STRING;
+		else if (strcasecmp(arg, "byte") == 0)
+			t |= UINT8;
+		else if (strcasecmp(arg, "uint16") == 0)
+			t |= UINT16;
+		else if (strcasecmp(arg, "int16") == 0)
+			t |= SINT16;
+		else if (strcasecmp(arg, "uint32") == 0)
+			t |= UINT32;
+		else if (strcasecmp(arg, "int32") == 0)
+			t |= SINT32;
+		else if (strcasecmp(arg, "flag") == 0)
+			t |= FLAG;
+		else if (strcasecmp(arg, "domain") == 0)
+			t |= STRING | RFC3397;
+		else if (strcasecmp(arg, "binhex") == 0)
+			t |= BINHEX;
+		else if (strcasecmp(arg, "embed") == 0)
+			t |= EMBED;
+		else if (strcasecmp(arg, "encap") == 0)
+			t |= ENCAP;
+		else if (strcasecmp(arg, "rfc3361") ==0)
+			t |= STRING | RFC3361;
+		else if (strcasecmp(arg, "rfc3442") ==0)
+			t |= STRING | RFC3442;
+		else if (strcasecmp(arg, "rfc5969") == 0)
+			t |= STRING | RFC5969;
+		else if (strcasecmp(arg, "option") == 0)
+			t |= OPTION;
+		else {
+			syslog(LOG_ERR, "unknown type: %s", arg);
+			return -1;
+		}
+		if (l && !(t & (STRING | BINHEX))) {
+			syslog(LOG_WARNING,
+			    "ignoring length for type `%s'", arg);
+			l = 0;
+		}
+		if (t & ARRAY && t & (STRING | BINHEX)) {
+			syslog(LOG_WARNING, "ignoring array for strings");
+			t &= ~ARRAY;
+		}
+		/* variable */
+		if (!fp) {
+			if (!(t & OPTION)) {
+			        syslog(LOG_ERR,
+				    "type %s requires a variable name", arg);
+				return -1;
+			}
+			np = NULL;
+		} else {
+			arg = strskipwhite(fp);
+			fp = strwhite(arg);
+			if (fp)
+				*fp++ = '\0';
+			np = strdup(arg);
+			if (np == NULL) {
+				syslog(LOG_ERR, "%s: %m", __func__);
+				return -1;
+			}
+		}
+		if (opt != O_EMBED) {
+			for (dl = 0, ndop = *dop; dl < *dop_len; dl++, ndop++)
+			{
+				/* type 0 seems freshly malloced struct
+				 * for us to use */
+				if (ndop->option == u || ndop->type == 0)
+					break;
+			}
+			if (dl == *dop_len)
+				ndop = NULL;
+		} else
+			ndop = NULL;
+		if (ndop == NULL) {
+			if ((ndop = realloc(*dop,
+			    sizeof(**dop) * ((*dop_len) + 1))) == NULL)
+			{
+				syslog(LOG_ERR, "%s: %m", __func__);
+				return -1;
+			}
+			*dop = ndop;
+			ndop = &(*dop)[(*dop_len)++];
+			ndop->embopts = NULL;
+			ndop->embopts_len = 0;
+			ndop->encopts = NULL;
+			ndop->encopts_len = 0;
+		} else
+			free_dhcp_opt_embenc(ndop);
+		ndop->option = u; /* could have been 0 */
+		ndop->type = t;
+		ndop->len = l;
+		ndop->var = np;
+		/* Save the define for embed and encap options */
+		if (opt == O_DEFINE || opt == O_DEFINE6 || opt == O_VENDOPT)
+			ldop = ndop;
+		else if (opt == O_ENCAP)
+			edop = ndop;
+		break;
+	case O_VENDCLASS:
+		fp = strwhite(arg);
+		if (fp)
+			*fp++ = '\0';
+		errno = 0;
+		u = strtoul(arg, &np, 0);
+		if (u > UINT32_MAX || errno != 0 || *np != '\0') {
+			syslog(LOG_ERR, "invalid code: %s", arg);
+			return -1;
+		}
+		if (fp) {
+			s = parse_string(NULL, 0, fp);
+			if (s == -1) {
+				syslog(LOG_ERR, "%s: %m", __func__);
+				return -1;
+			}
+			if (s + (sizeof(uint16_t) * 2) > UINT16_MAX) {
+				syslog(LOG_ERR, "vendor class is too big");
+				return -1;
+			}
+			np = malloc(s);
+			if (np == NULL) {
+				syslog(LOG_ERR, "%s: %m", __func__);
+				return -1;
+			}
+			parse_string(np, s, fp);
+		} else {
+			s = 0;
+			np = NULL;
+		}
+		vivco = realloc(ifo->vivco, sizeof(*ifo->vivco) *
+		    (ifo->vivco_len + 1));
+		if (vivco == NULL) {
+			syslog(LOG_ERR, "%s: %m", __func__);
+			return -1;
+		}
+		ifo->vivco = vivco;
+		ifo->vivco_en = u;
+		vivco = &ifo->vivco[ifo->vivco_len++];
+		vivco->len = s;
+		vivco->data = (uint8_t *)np;
+		break;
 	default:
 		return 0;
 	}
@@ -1146,42 +1531,13 @@ finish_config(struct if_options *ifo)
 	if (ifo->vendor[0] && !(ifo->options & DHCPCD_VENDORRAW)) {
 		ifo->vendor[0]++;
 		ifo->vendor[ifo->vendor[0]] = DHO_END;
+		/* We are called twice.
+		 * This should be fixed, but in the meantime, this
+		 * guard should suffice */
+		ifo->options |= DHCPCD_VENDORRAW;
 	}
 }
 
-#ifdef INET6
-static void
-finish_config6(struct if_options *ifo, const char *ifname)
-{
-
-	if (!(ifo->options & DHCPCD_IPV6))
-		ifo->options &= ~DHCPCD_IPV6RS;
-
-	if (ifname && ifo->iaid_len == 0 && ifo->options & DHCPCD_IPV6) {
-		ifo->iaid = malloc(sizeof(*ifo->iaid));
-		if (ifo->iaid == NULL)
-			syslog(LOG_ERR, "%s: %m", __func__);
-		else {
-			if (ifo->ia_type == 0)
-				ifo->ia_type = D6_OPTION_IA_NA;
-			ifo->iaid_len = strlen(ifname);
-			if (ifo->iaid_len <= sizeof(ifo->iaid->iaid)) {
-				strncpy((char *)ifo->iaid->iaid, ifname,
-					sizeof(ifo->iaid->iaid));
-				memset(ifo->iaid->iaid + ifo->iaid_len, 0,
-					sizeof(ifo->iaid->iaid) -ifo->iaid_len);
-			} else {
-				uint32_t idx = if_nametoindex(ifname);
-				memcpy(ifo->iaid->iaid, &idx, sizeof(idx));
-			}
-			ifo->iaid_len = 1;
-			ifo->iaid->sla = NULL;
-			ifo->iaid->sla_len = 0;
-		}
-	}
-}
-#endif
-
 struct if_options *
 read_config(const char *file,
     const char *ifname, const char *ssid, const char *profile)
@@ -1190,6 +1546,14 @@ read_config(const char *file,
 	FILE *f;
 	char *line, *option, *p;
 	int skip = 0, have_profile = 0;
+#ifndef EMBEDDED_CONFIG
+	char *buf;
+	const char **e;
+	size_t buflen, ol;
+#endif
+#if !defined(INET) || !defined(INET6)
+	struct dhcp_opt *opt;
+#endif
 
 	/* Seed our default options */
 	ifo = calloc(1, sizeof(*ifo));
@@ -1217,6 +1581,108 @@ read_config(const char *file,
 	ifo->vendorclassid[0] = strlen(vendor);
 	memcpy(ifo->vendorclassid + 1, vendor, ifo->vendorclassid[0]);
 
+	/* Parse our embedded options file */
+	if (ifname == NULL) {
+		/* Space for initial estimates */
+#if defined(INET) && defined(INITDEFINES)
+		ifo->dhcp_override =
+		    calloc(INITDEFINES, sizeof(*ifo->dhcp_override));
+		if (ifo->dhcp_override == NULL)
+			syslog(LOG_ERR, "%s: %m", __func__);
+		else
+			ifo->dhcp_override_len = INITDEFINES;
+#endif
+
+#if defined(INET6) && defined(INITDEFINE6S)
+		ifo->dhcp6_override =
+		    calloc(INITDEFINE6S, sizeof(*ifo->dhcp6_override));
+		if (ifo->dhcp6_override == NULL)
+			syslog(LOG_ERR, "%s: %m", __func__);
+		else
+			ifo->dhcp6_override_len = INITDEFINE6S;
+#endif
+
+		/* Now load our embedded config */
+#ifdef EMBEDDED_CONFIG
+		f = fopen(EMBEDDED_CONFIG, "r");
+		if (f == NULL)
+			syslog(LOG_ERR, "fopen `%s': %m", EMBEDDED_CONFIG);
+
+		while (f && (line = get_line(f))) {
+#else
+		buflen = 80;
+		buf = malloc(buflen);
+		if (buf == NULL) {
+			syslog(LOG_ERR, "%s: %m", __func__);
+			return NULL;
+		}
+		for (e = dhcpcd_embedded_conf; *e; e++) {
+			ol = strlen(*e) + 1;
+			if (ol > buflen) {
+				free(buf);
+				buflen = ol;
+				buf = malloc(buflen);
+				if (buf == NULL) {
+					syslog(LOG_ERR, "%s: %m", __func__);
+					return NULL;
+				}
+			}
+			memcpy(buf, *e, ol);
+			line = buf;
+#endif
+			option = strsep(&line, " \t");
+			if (line)
+				line = strskipwhite(line);
+			/* Trim trailing whitespace */
+			if (line && *line) {
+				p = line + strlen(line) - 1;
+				while (p != line &&
+				    (*p == ' ' || *p == '\t') &&
+				    *(p - 1) != '\\')
+					*p-- = '\0';
+			}
+			parse_config_line(ifo, option, line);
+
+		}
+
+#ifdef EMBEDDED_CONFIG
+		if (f)
+			fclose(f);
+#else
+		free(buf);
+#endif
+#ifdef INET
+		dhcp_opts = ifo->dhcp_override;
+		dhcp_opts_len = ifo->dhcp_override_len;
+#else
+		for (i = 0, opt = ifo->dhcp_override;
+		    i < ifo->dhcp_override_len;
+		    i++, opt++)
+			free_dhcp_opt_embenc(opt);
+		free(ifo->dhcp_override);
+#endif
+		ifo->dhcp_override = NULL;
+		ifo->dhcp_override_len = 0;
+
+#ifdef INET6
+		dhcp6_opts = ifo->dhcp6_override;
+		dhcp6_opts_len = ifo->dhcp6_override_len;
+#else
+		for (i = 0, opt = ifo->dhcp6_override;
+		    i < ifo->dhcp_override6_len;
+		    i++, opt++)
+			free_dhcp_opt_embenc(opt);
+		free(ifo->dhcp6_override);
+#endif
+		ifo->dhcp6_override = NULL;
+		ifo->dhcp6_override_len = 0;
+
+		vivso = ifo->vivso_override;
+		vivso_len = ifo->vivso_override_len;
+		ifo->vivso_override = NULL;
+		ifo->vivso_override_len = 0;
+	}
+
 	/* Parse our options file */
 	f = fopen(file ? file : CONFIG, "r");
 	if (f == NULL) {
@@ -1227,6 +1693,8 @@ read_config(const char *file,
 
 	while ((line = get_line(f))) {
 		option = strsep(&line, " \t");
+		if (line)
+			line = strskipwhite(line);
 		/* Trim trailing whitespace */
 		if (line && *line) {
 			p = line + strlen(line) - 1;
@@ -1273,9 +1741,6 @@ read_config(const char *file,
 	}
 
 	finish_config(ifo);
-#ifdef INET6
-	finish_config6(ifo, ifname);
-#endif
 	return ifo;
 }
 
@@ -1297,9 +1762,6 @@ add_options(struct if_options *ifo, int argc, char **argv)
 	}
 
 	finish_config(ifo);
-#ifdef INET6
-	finish_config6(ifo, NULL);
-#endif
 	return r;
 }
 
@@ -1307,6 +1769,8 @@ void
 free_options(struct if_options *ifo)
 {
 	size_t i;
+	struct dhcp_opt *opt;
+	struct vivco *vo;
 
 	if (ifo) {
 		if (ifo->environ) {
@@ -1325,11 +1789,34 @@ free_options(struct if_options *ifo)
 		free(ifo->arping);
 		free(ifo->blacklist);
 		free(ifo->fallback);
+
+		for (i = 0, opt = ifo->dhcp_override;
+		    i < ifo->dhcp_override_len;
+		    i++, opt++)
+			free_dhcp_opt_embenc(opt);
+		free(ifo->dhcp_override);
+		for (i = 0, opt = ifo->dhcp6_override;
+		    i < ifo->dhcp6_override_len;
+		    i++, opt++)
+			free_dhcp_opt_embenc(opt);
+		free(ifo->dhcp6_override);
+		for (i = 0, vo = ifo->vivco;
+		    i < ifo->vivco_len;
+		    i++, vo++)
+			free(vo->data);
+		free(ifo->vivco);
+		for (i = 0, opt = ifo->vivso_override;
+		    i < ifo->vivso_override_len;
+		    i++, opt++)
+			free_dhcp_opt_embenc(opt);
+		free(ifo->vivso_override);
+
 #ifdef INET6
-		for (i = 0; i < ifo->iaid_len; i++)
-			free(ifo->iaid[i].sla);
-		free(ifo->iaid);
+		for (i = 0; i < ifo->ia_len; i++)
+			free(ifo->ia[i].sla);
 #endif
+		free(ifo->ia);
+
 		free(ifo);
 	}
 }
