@@ -1,6 +1,6 @@
-/* 
+/*
  * dhcpcd - DHCP client daemon
- * Copyright (c) 2006-2011 Roy Marples <roy@marples.name>
+ * Copyright (c) 2006-2013 Roy Marples <roy@marples.name>
  * All rights reserved
 
  * Redistribution and use in source and binary forms, with or without
@@ -25,25 +25,22 @@
  * SUCH DAMAGE.
  */
 
+#include <sys/param.h>
 #include <sys/types.h>
 #include <sys/ioctl.h>
-#include <sys/param.h>
 #include <sys/socket.h>
-#include <sys/time.h>
 
-#include <arpa/inet.h>
 #include <net/if.h>
 #include <net/if_arp.h>
+#include <netinet/in.h>
+#ifdef __FreeBSD__ /* Needed so that including netinet6/in6_var.h works */
+#  include <net/if_var.h>
+#endif
 #ifdef AF_LINK
 #  include <net/if_dl.h>
 #  include <net/if_types.h>
+#  include <netinet/in_var.h>
 #endif
-#include <netinet/in_systm.h>
-#include <netinet/in.h>
-#include <netinet/ip.h>
-#define __FAVOR_BSD /* Nasty glibc hack so we can use BSD semantics for UDP */
-#include <netinet/udp.h>
-#undef __FAVOR_BSD
 #ifdef AF_PACKET
 #  include <netpacket/packet.h>
 #endif
@@ -51,6 +48,12 @@
 #  include <net/if_media.h>
 #endif
 
+#include <net/route.h>
+#ifdef __linux__
+#  include <asm/types.h> /* for systems with broken headers */
+#  include <linux/rtnetlink.h>
+#endif
+
 #include <ctype.h>
 #include <errno.h>
 #include <ifaddrs.h>
@@ -64,68 +67,17 @@
 
 #include "config.h"
 #include "common.h"
+#include "dev.h"
 #include "dhcp.h"
+#include "dhcp6.h"
 #include "if-options.h"
-#include "ipv6rs.h"
+#include "ipv4.h"
+#include "ipv6nd.h"
 #include "net.h"
-#include "signals.h"
-
-static char hwaddr_buffer[(HWADDR_LEN * 3) + 1 + 1024];
 
 int socket_afnet = -1;
 
-int
-inet_ntocidr(struct in_addr address)
-{
-	int cidr = 0;
-	uint32_t mask = htonl(address.s_addr);
-
-	while (mask) {
-		cidr++;
-		mask <<= 1;
-	}
-	return cidr;
-}
-
-int
-inet_cidrtoaddr(int cidr, struct in_addr *addr)
-{
-	int ocets;
-
-	if (cidr < 1 || cidr > 32) {
-		errno = EINVAL;
-		return -1;
-	}
-	ocets = (cidr + 7) / 8;
-
-	addr->s_addr = 0;
-	if (ocets > 0) {
-		memset(&addr->s_addr, 255, (size_t)ocets - 1);
-		memset((unsigned char *)&addr->s_addr + (ocets - 1),
-		    (256 - (1 << (32 - cidr) % 8)), 1);
-	}
-
-	return 0;
-}
-
-uint32_t
-get_netmask(uint32_t addr)
-{
-	uint32_t dst;
-
-	if (addr == 0)
-		return 0;
-
-	dst = htonl(addr);
-	if (IN_CLASSA(dst))
-		return ntohl(IN_CLASSA_NET);
-	if (IN_CLASSB(dst))
-		return ntohl(IN_CLASSB_NET);
-	if (IN_CLASSC(dst))
-		return ntohl(IN_CLASSC_NET);
-
-	return 0;
-}
+static char hwaddr_buffer[(HWADDR_LEN * 3) + 1 + 1024];
 
 char *
 hwaddr_ntoa(const unsigned char *hwaddr, size_t hwlen)
@@ -183,21 +135,18 @@ hwaddr_aton(unsigned char *buffer, const char *addr)
 }
 
 void
-free_interface(struct interface *iface)
+free_interface(struct interface *ifp)
 {
-	if (!iface)
+
+	if (ifp == NULL)
 		return;
-	ipv6rs_free(iface);
-	if (iface->state) {
-		free_options(iface->state->options);
-		free(iface->state->old);
-		free(iface->state->new);
-		free(iface->state->offer);
-		free(iface->state);
-	}
-	free(iface->buffer);
-	free(iface->clientid);
-	free(iface);
+	ipv4_free(ifp);
+	dhcp_free(ifp);
+	ipv6_free(ifp);
+	dhcp6_free(ifp);
+	ipv6nd_free(ifp);
+	free_options(ifp->options);
+	free(ifp);
 }
 
 int
@@ -221,19 +170,19 @@ carrier_status(struct interface *iface)
 #endif
 
 	if (ioctl(socket_afnet, SIOCGIFFLAGS, &ifr) == -1)
-		return -1;
+		return LINK_UNKNOWN;
 	iface->flags = ifr.ifr_flags;
 
-	ret = -1;
+	ret = LINK_UNKNOWN;
 #ifdef SIOCGIFMEDIA
 	memset(&ifmr, 0, sizeof(ifmr));
 	strlcpy(ifmr.ifm_name, iface->name, sizeof(ifmr.ifm_name));
 	if (ioctl(socket_afnet, SIOCGIFMEDIA, &ifmr) != -1 &&
 	    ifmr.ifm_status & IFM_AVALID)
-		ret = (ifmr.ifm_status & IFM_ACTIVE) ? 1 : 0;
+		ret = (ifmr.ifm_status & IFM_ACTIVE) ? LINK_UP : LINK_DOWN;
 #endif
-	if (ret == -1)
-		ret = (ifr.ifr_flags & IFF_RUNNING) ? 1 : 0;
+	if (ret == LINK_UNKNOWN)
+		ret = (ifr.ifr_flags & IFF_RUNNING) ? LINK_UP : LINK_DOWN;
 	return ret;
 }
 
@@ -266,16 +215,26 @@ up_interface(struct interface *iface)
 	return retval;
 }
 
-struct interface *
+struct if_head *
 discover_interfaces(int argc, char * const *argv)
 {
 	struct ifaddrs *ifaddrs, *ifa;
 	char *p;
 	int i, sdl_type;
-	struct interface *ifp, *ifs, *ifl;
+	struct if_head *ifs;
+	struct interface *ifp;
 #ifdef __linux__
 	char ifn[IF_NAMESIZE];
 #endif
+#ifdef INET
+	const struct sockaddr_in *addr;
+	const struct sockaddr_in *net;
+	const struct sockaddr_in *dst;
+#endif
+#ifdef INET6
+	const struct sockaddr_in6 *sin6;
+	int ifa_flags;
+#endif
 #ifdef AF_LINK
 	const struct sockaddr_dl *sdl;
 #ifdef IFLR_ACTIVE
@@ -294,7 +253,11 @@ discover_interfaces(int argc, char * const *argv)
 	if (getifaddrs(&ifaddrs) == -1)
 		return NULL;
 
-	ifs = ifl = NULL;
+	ifs = malloc(sizeof(*ifs));
+	if (ifs == NULL)
+		return NULL;
+	TAILQ_INIT(ifs);
+
 	for (ifa = ifaddrs; ifa; ifa = ifa->ifa_next) {
 		if (ifa->ifa_addr != NULL) {
 #ifdef AF_LINK
@@ -306,11 +269,16 @@ discover_interfaces(int argc, char * const *argv)
 #endif
 		}
 
+		/* Ensure that the interface name has settled */
+		if (!dev_initialized(ifa->ifa_name))
+			continue;
+
 		/* It's possible for an interface to have >1 AF_LINK.
 		 * For our purposes, we use the first one. */
-		for (ifp = ifs; ifp; ifp = ifp->next)
+		TAILQ_FOREACH(ifp, ifs, next) {
 			if (strcmp(ifp->name, ifa->ifa_name) == 0)
 				break;
+		}
 		if (ifp)
 			continue;
 		if (argc > 0) {
@@ -332,39 +300,42 @@ discover_interfaces(int argc, char * const *argv)
 				continue;
 			p = argv[i];
 		} else {
+			p = ifa->ifa_name;
 			/* -1 means we're discovering against a specific
 			 * interface, but we still need the below rules
 			 * to apply. */
 			if (argc == -1 && strcmp(argv[0], ifa->ifa_name) != 0)
 				continue;
-			for (i = 0; i < ifdc; i++)
-				if (!fnmatch(ifdv[i], ifa->ifa_name, 0))
-					break;
-			if (i < ifdc)
-				continue;
-			for (i = 0; i < ifac; i++)
-				if (!fnmatch(ifav[i], ifa->ifa_name, 0))
-					break;
-			if (ifac && i == ifac)
-				continue;
-			p = ifa->ifa_name;
 		}
+		for (i = 0; i < ifdc; i++)
+			if (!fnmatch(ifdv[i], p, 0))
+				break;
+		if (i < ifdc)
+			continue;
+		for (i = 0; i < ifac; i++)
+			if (!fnmatch(ifav[i], p, 0))
+				break;
+		if (ifac && i == ifac)
+			continue;
 
-		ifp = xzalloc(sizeof(*ifp));
+		ifp = calloc(1, sizeof(*ifp));
+		if (ifp == NULL)
+			return NULL;
 		strlcpy(ifp->name, p, sizeof(ifp->name));
 		ifp->flags = ifa->ifa_flags;
 
 		/* Bring the interface up if not already */
 		if (!(ifp->flags & IFF_UP)
 #ifdef SIOCGIFMEDIA
-		    && carrier_status(ifp) != -1
+		    && carrier_status(ifp) != LINK_UNKNOWN
 #endif
 		   )
 		{
 			if (up_interface(ifp) == 0)
 				options |= DHCPCD_WAITUP;
 			else
-				syslog(LOG_ERR, "%s: up_interface: %m", ifp->name);
+				syslog(LOG_ERR, "%s: up_interface: %m",
+				    ifp->name);
 		}
 
 		sdl_type = 0;
@@ -426,6 +397,11 @@ discover_interfaces(int argc, char * const *argv)
 				memcpy(ifp->hwaddr, sll->sll_addr, ifp->hwlen);
 #endif
 		}
+#ifdef __linux__
+		/* PPP addresses on Linux don't have hardware addresses */
+		else
+			ifp->index = if_nametoindex(ifp->name);
+#endif
 
 		/* We only work on ethernet by default */
 		if (!(ifp->flags & IFF_POINTOPOINT) &&
@@ -473,17 +449,46 @@ discover_interfaces(int argc, char * const *argv)
 			ifp->wireless = 1;
 			ifp->metric += 100;
 		}
-		snprintf(ifp->leasefile, sizeof(ifp->leasefile),
-		    LEASEFILE, ifp->name);
-		/* 0 is a valid fd, so init to -1 */
-		ifp->raw_fd = ifp->udp_fd = ifp->arp_fd = -1;
 
-		if (ifl)
-			ifl->next = ifp; 
-		else
-			ifs = ifp;
-		ifl = ifp;
+		TAILQ_INSERT_TAIL(ifs, ifp, next);
 	}
+
+	for (ifa = ifaddrs; ifa; ifa = ifa->ifa_next) {
+		if (ifa->ifa_addr == NULL)
+			continue;
+		switch(ifa->ifa_addr->sa_family) {
+#ifdef INET
+		case AF_INET:
+			addr = (const struct sockaddr_in *)
+			    (void *)ifa->ifa_addr;
+			net = (const struct sockaddr_in *)
+			    (void *)ifa->ifa_netmask;
+			if (ifa->ifa_flags & IFF_POINTOPOINT)
+				dst = (const struct sockaddr_in *)
+				    (void *)ifa->ifa_dstaddr;
+			else
+				dst = NULL;
+			ipv4_handleifa(RTM_NEWADDR, ifs, ifa->ifa_name,
+				&addr->sin_addr,
+				&net->sin_addr,
+				dst ? &dst->sin_addr : NULL);
+			break;
+#endif
+#ifdef INET6
+		case AF_INET6:
+			sin6 = (const struct sockaddr_in6 *)
+			    (void *)ifa->ifa_addr;
+			ifa_flags = in6_addr_flags(ifa->ifa_name,
+			    &sin6->sin6_addr);
+			if (ifa_flags != -1)
+				ipv6_handleifa(RTM_NEWADDR, ifs,
+				    ifa->ifa_name,
+				    &sin6->sin6_addr, ifa_flags);
+			break;
+#endif
+		}
+	}
+
 	freeifaddrs(ifaddrs);
 
 #ifdef IFLR_ACTIVE
@@ -493,53 +498,6 @@ discover_interfaces(int argc, char * const *argv)
 	return ifs;
 }
 
-int
-do_address(const char *ifname,
-    struct in_addr *addr, struct in_addr *net, struct in_addr *dst, int act)
-{
-	struct ifaddrs *ifaddrs, *ifa;
-	const struct sockaddr_in *a, *n, *d;
-	int retval;
-
-	if (getifaddrs(&ifaddrs) == -1)
-		return -1;
-
-	retval = 0;
-	for (ifa = ifaddrs; ifa; ifa = ifa->ifa_next) {
-		if (ifa->ifa_addr == NULL ||
-		    ifa->ifa_addr->sa_family != AF_INET ||
-		    strcmp(ifa->ifa_name, ifname) != 0)
-			continue;
-		a = (const struct sockaddr_in *)(void *)ifa->ifa_addr;
-		n = (const struct sockaddr_in *)(void *)ifa->ifa_netmask;
-		if (ifa->ifa_flags & IFF_POINTOPOINT)
-			d = (const struct sockaddr_in *)(void *)
-			    ifa->ifa_dstaddr;
-		else
-			d = NULL;
-		if (act == 1) {
-			addr->s_addr = a->sin_addr.s_addr;
-			net->s_addr = n->sin_addr.s_addr;
-			if (dst) {
-				if (ifa->ifa_flags & IFF_POINTOPOINT)
-					dst->s_addr = d->sin_addr.s_addr;
-				else
-					dst->s_addr = INADDR_ANY;
-			}
-			retval = 1;
-			break;
-		}
-		if (addr->s_addr == a->sin_addr.s_addr &&
-		    (net == NULL || net->s_addr == n->sin_addr.s_addr))
-		{
-			retval = 1;
-			break;
-		}
-	}
-	freeifaddrs(ifaddrs);
-	return retval;
-}
-
 int
 do_mtu(const char *ifname, short int mtu)
 {
@@ -554,219 +512,3 @@ do_mtu(const char *ifname, short int mtu)
 		return -1;
 	return ifr.ifr_mtu;
 }
-
-void
-free_routes(struct rt *routes)
-{
-	struct rt *r;
-
-	while (routes) {
-		r = routes->next;
-		free(routes);
-		routes = r;
-	}
-}
-
-int
-open_udp_socket(struct interface *iface)
-{
-	int s;
-	struct sockaddr_in sin;
-	int n;
-#ifdef SO_BINDTODEVICE
-	struct ifreq ifr;
-	char *p;
-#endif
-
-	if ((s = socket(PF_INET, SOCK_DGRAM, IPPROTO_UDP)) == -1)
-		return -1;
-
-	n = 1;
-	if (setsockopt(s, SOL_SOCKET, SO_REUSEADDR, &n, sizeof(n)) == -1)
-		goto eexit;
-#ifdef SO_BINDTODEVICE
-	memset(&ifr, 0, sizeof(ifr));
-	strlcpy(ifr.ifr_name, iface->name, sizeof(ifr.ifr_name));
-	/* We can only bind to the real device */
-	p = strchr(ifr.ifr_name, ':');
-	if (p)
-		*p = '\0';
-	if (setsockopt(s, SOL_SOCKET, SO_BINDTODEVICE, &ifr,
-		sizeof(ifr)) == -1)
-		goto eexit;
-#endif
-	/* As we don't use this socket for receiving, set the
-	 * receive buffer to 1 */
-	n = 1;
-	if (setsockopt(s, SOL_SOCKET, SO_RCVBUF, &n, sizeof(n)) == -1)
-		goto eexit;
-	memset(&sin, 0, sizeof(sin));
-	sin.sin_family = AF_INET;
-	sin.sin_port = htons(DHCP_CLIENT_PORT);
-	sin.sin_addr.s_addr = iface->addr.s_addr;
-	if (bind(s, (struct sockaddr *)&sin, sizeof(sin)) == -1)
-		goto eexit;
-
-	iface->udp_fd = s;
-	set_cloexec(s);
-	return 0;
-
-eexit:
-	close(s);
-	return -1;
-}
-
-ssize_t
-send_packet(const struct interface *iface, struct in_addr to,
-    const uint8_t *data, ssize_t len)
-{
-	struct sockaddr_in sin;
-
-	memset(&sin, 0, sizeof(sin));
-	sin.sin_family = AF_INET;
-	sin.sin_addr.s_addr = to.s_addr;
-	sin.sin_port = htons(DHCP_SERVER_PORT);
-	return sendto(iface->udp_fd, data, len, 0,
-	    (struct sockaddr *)&sin, sizeof(sin));
-}
-
-struct udp_dhcp_packet
-{
-	struct ip ip;
-	struct udphdr udp;
-	struct dhcp_message dhcp;
-};
-const size_t udp_dhcp_len = sizeof(struct udp_dhcp_packet);
-
-static uint16_t
-checksum(const void *data, uint16_t len)
-{
-	const uint8_t *addr = data;
-	uint32_t sum = 0;
-
-	while (len > 1) {
-		sum += addr[0] * 256 + addr[1];
-		addr += 2;
-		len -= 2;
-	}
-
-	if (len == 1)
-		sum += *addr * 256;
-
-	sum = (sum >> 16) + (sum & 0xffff);
-	sum += (sum >> 16);
-
-	sum = htons(sum);
-
-	return ~sum;
-}
-
-ssize_t
-make_udp_packet(uint8_t **packet, const uint8_t *data, size_t length,
-    struct in_addr source, struct in_addr dest)
-{
-	struct udp_dhcp_packet *udpp;
-	struct ip *ip;
-	struct udphdr *udp;
-
-	udpp = xzalloc(sizeof(*udpp));
-	ip = &udpp->ip;
-	udp = &udpp->udp;
-
-	/* OK, this is important :)
-	 * We copy the data to our packet and then create a small part of the
-	 * ip structure and an invalid ip_len (basically udp length).
-	 * We then fill the udp structure and put the checksum
-	 * of the whole packet into the udp checksum.
-	 * Finally we complete the ip structure and ip checksum.
-	 * If we don't do the ordering like so then the udp checksum will be
-	 * broken, so find another way of doing it! */
-
-	memcpy(&udpp->dhcp, data, length);
-
-	ip->ip_p = IPPROTO_UDP;
-	ip->ip_src.s_addr = source.s_addr;
-	if (dest.s_addr == 0)
-		ip->ip_dst.s_addr = INADDR_BROADCAST;
-	else
-		ip->ip_dst.s_addr = dest.s_addr;
-
-	udp->uh_sport = htons(DHCP_CLIENT_PORT);
-	udp->uh_dport = htons(DHCP_SERVER_PORT);
-	udp->uh_ulen = htons(sizeof(*udp) + length);
-	ip->ip_len = udp->uh_ulen;
-	udp->uh_sum = checksum(udpp, sizeof(*udpp));
-
-	ip->ip_v = IPVERSION;
-	ip->ip_hl = sizeof(*ip) >> 2;
-	ip->ip_id = arc4random() & UINT16_MAX;
-	ip->ip_ttl = IPDEFTTL;
-	ip->ip_len = htons(sizeof(*ip) + sizeof(*udp) + length);
-	ip->ip_sum = checksum(ip, sizeof(*ip));
-
-	*packet = (uint8_t *)udpp;
-	return sizeof(*ip) + sizeof(*udp) + length;
-}
-
-ssize_t
-get_udp_data(const uint8_t **data, const uint8_t *udp)
-{
-	struct udp_dhcp_packet packet;
-
-	memcpy(&packet, udp, sizeof(packet));
-	*data = udp + offsetof(struct udp_dhcp_packet, dhcp);
-	return ntohs(packet.ip.ip_len) -
-	    sizeof(packet.ip) -
-	    sizeof(packet.udp);
-}
-
-int
-valid_udp_packet(const uint8_t *data, size_t data_len, struct in_addr *from,
-    int noudpcsum)
-{
-	struct udp_dhcp_packet packet;
-	uint16_t bytes, udpsum;
-
-	if (data_len < sizeof(packet.ip)) {
-		if (from)
-			from->s_addr = INADDR_ANY;
-		errno = EINVAL;
-		return -1;
-	}
-	memcpy(&packet, data, MIN(data_len, sizeof(packet)));
-	if (from)
-		from->s_addr = packet.ip.ip_src.s_addr;
-	if (data_len > sizeof(packet)) {
-		errno = EINVAL;
-		return -1;
-	}
-	if (checksum(&packet.ip, sizeof(packet.ip)) != 0) {
-		errno = EINVAL;
-		return -1;
-	}
-
-	bytes = ntohs(packet.ip.ip_len);
-	if (data_len < bytes) {
-		errno = EINVAL;
-		return -1;
-	}
-
-	if (noudpcsum == 0) {
-		udpsum = packet.udp.uh_sum;
-		packet.udp.uh_sum = 0;
-		packet.ip.ip_hl = 0;
-		packet.ip.ip_v = 0;
-		packet.ip.ip_tos = 0;
-		packet.ip.ip_len = packet.udp.uh_ulen;
-		packet.ip.ip_id = 0;
-		packet.ip.ip_off = 0;
-		packet.ip.ip_ttl = 0;
-		packet.ip.ip_sum = 0;
-		if (udpsum && checksum(&packet, bytes) != udpsum) {
-			errno = EINVAL;
-			return -1;
-		}
-	}
-
-	return 0;
-}
