@@ -421,7 +421,7 @@ dumplease:
 		}
 		elen += e;
 	}
-	env[elen] = '\0';
+	env[elen] = NULL;
 
 	*argv = env;
 	return elen;
@@ -553,7 +553,7 @@ script_runreason(const struct interface *ifp, const char *reason)
 			goto out;
 		}
 	}
-	env[++elen] = '\0';
+	env[++elen] = NULL;
 
 	pid = exec_script(argv, env);
 	if (pid == -1)
